import json
import os
from pathlib import Path
import shutil
import subprocess
import warnings

from PIL import Image, ImageChops
import pytest

from pelican.plugins.image_process import (
    ExifTool,
    compute_paths,
    harvest_images_in_fragment,
    is_img_identifiable,
    process_image,
)

# Prepare test image constants.
HERE = Path(__file__).resolve().parent
TEST_DATA = HERE.joinpath("test_data").resolve()
TEST_IMAGES = [
    TEST_DATA.joinpath(f"pelican-bird.{ext}").resolve() for ext in ["jpg", "png"]
]
EXIF_TEST_IMAGES = [
    TEST_DATA.joinpath("exif", f"pelican-bird.{ext}").resolve()
    for ext in ["jpg", "png"]
]
TRANSFORM_RESULTS = TEST_DATA.joinpath("results").resolve()

# Register all supported transforms.
SINGLE_TRANSFORMS = {
    "crop": ["crop 10 20 100 200"],
    "flip_horizontal": ["flip_horizontal"],
    "flip_vertical": ["flip_vertical"],
    "grayscale": ["grayscale"],
    "resize": ["resize 200 250"],
    "rotate": ["rotate 20"],
    "scale_in": ["scale_in 200 250 False"],
    "scale_out": ["scale_out 200 250 False"],
    "blur": ["blur"],
    "contour": ["contour"],
    "detail": ["detail"],
    "edge_enhance": ["edge_enhance"],
    "edge_enhance_more": ["edge_enhance_more"],
    "emboss": ["emboss"],
    "find_edges": ["find_edges"],
    "smooth": ["smooth"],
    "smooth_more": ["smooth_more"],
    "sharpen": ["sharpen"],
}


def get_settings(**kwargs):
    """Provide tweaked setting dictionaries for testing
    Set keyword arguments to override specific settings.
    """
    DEFAULT_CONFIG = {
        "PATH": HERE,
        "OUTPUT_PATH": "output",
        "static_content": {},
        "filenames": {},
        "SITEURL": "//",
        "IMAGE_PROCESS": SINGLE_TRANSFORMS,
    }
    settings = DEFAULT_CONFIG.copy()
    settings.update(kwargs)
    return settings


def test_undefined_transform():
    settings = get_settings()
    tag = "<img class='image-process-undefined' src='/tmp/test.jpg' />"
    with pytest.raises(RuntimeError):
        harvest_images_in_fragment(tag, settings)


@pytest.mark.parametrize("transform_id, transform_params", SINGLE_TRANSFORMS.items())
@pytest.mark.parametrize("image_path", TEST_IMAGES)
def test_all_transforms(tmp_path, transform_id, transform_params, image_path):
    """Test the raw transform and their results on images."""
    settings = get_settings()

    image_name = image_path.name
    destination_path = tmp_path.joinpath(transform_id, image_name)
    expected_path = TRANSFORM_RESULTS.joinpath(transform_id, image_name)

    process_image((str(image_path), str(destination_path), transform_params), settings)

    transformed = Image.open(destination_path)
    expected = Image.open(expected_path)
    # Image.getbbox() returns None if there are only black pixels in the image:
    image_diff = ImageChops.difference(transformed, expected).getbbox()
    assert image_diff is None


@pytest.mark.parametrize(
    "orig_src, orig_img, new_src, new_img",
    [
        (
            "/tmp/test.jpg",
            "tmp/test.jpg",
            "/tmp/derivatives/crop/test.jpg",
            "tmp/derivatives/crop/test.jpg",
        ),
        (
            "../tmp/test.jpg",
            "../tmp/test.jpg",
            "../tmp/derivatives/crop/test.jpg",
            "../tmp/derivatives/crop/test.jpg",
        ),
        (
            "http://xxx/tmp/test.jpg",
            "tmp/test.jpg",
            "http://xxx/tmp/derivatives/crop/test.jpg",
            "tmp/derivatives/crop/test.jpg",
        ),
    ],
)
def test_path_normalization(mocker, orig_src, orig_img, new_src, new_img):
    # Allow non-existing images to be processed:
    mocker.patch(
        "pelican.plugins.image_process.image_process.is_img_identifiable",
        lambda img_filepath: True,
    )
    # Silence image transforms.
    process = mocker.patch("pelican.plugins.image_process.image_process.process_image")

    settings = get_settings(IMAGE_PROCESS_DIR="derivatives")

    img_tag_orig = (
        '<img class="test image-process image-process-crop test2" '
        f'src="{orig_src}"/>'
    )

    img_tag_processed = harvest_images_in_fragment(img_tag_orig, settings)

    assert img_tag_processed == (
        f'<img class="test image-process image-process-crop test2" src="{new_src}"/>'
    )

    process.assert_called_once_with(
        (
            os.path.join(settings["PATH"], orig_img),
            os.path.join(settings["OUTPUT_PATH"], new_img),
            SINGLE_TRANSFORMS["crop"],
        ),
        settings,
    )


COMPLEX_TRANSFORMS = {
    "thumb": ["crop 0 0 50% 50%", "scale_out 150 150", "crop 0 0 150 150"],
    "article-image": {"type": "image", "ops": ["scale_in 300 300"]},
    "crisp": {
        "type": "responsive-image",
        "srcset": [
            ("1x", ["scale_in 800 600 True"]),
            ("2x", ["scale_in 1600 1200 True"]),
            ("4x", ["scale_in 3200 2400 True"]),
        ],
        "default": "1x",
    },
    "crisp2": {
        "type": "responsive-image",
        "srcset": [
            ("1x", ["scale_in 800 600 True"]),
            ("2x", ["scale_in 1600 1200 True"]),
            ("4x", ["scale_in 3200 2400 True"]),
        ],
        "default": ["scale_in 400 300 True"],
    },
    "large-photo": {
        "type": "responsive-image",
        "sizes": (
            "(min-width: 1200px) 800px, "
            "(min-width: 992px) 650px, "
            "(min-width: 768px) 718px, "
            "100vw"
        ),
        "srcset": [
            ("600w", ["scale_in 600 450 True"]),
            ("800w", ["scale_in 800 600 True"]),
            ("1600w", ["scale_in 1600 1200 True"]),
        ],
        "default": "800w",
    },
    "pict": {
        "type": "picture",
        "sources": [
            {
                "name": "default",
                "media": "(min-width: 640px)",
                "srcset": [
                    ("640w", ["scale_in 640 480 True"]),
                    ("1024w", ["scale_in 1024 683 True"]),
                    ("1600w", ["scale_in 1600 1200 True"]),
                ],
                "sizes": "100vw",
            },
            {
                "name": "source-1",
                "srcset": [
                    ("1x", ["crop 100 100 200 200"]),
                    ("2x", ["crop 100 100 300 300"]),
                ],
            },
        ],
        "default": ("default", "640w"),
    },
    "pict2": {
        "type": "picture",
        "sources": [
            {
                "name": "default",
                "media": "(min-width: 640px)",
                "srcset": [
                    ("640w", ["scale_in 640 480 True"]),
                    ("1024w", ["scale_in 1024 683 True"]),
                    ("1600w", ["scale_in 1600 1200 True"]),
                ],
                "sizes": "100vw",
            },
            {
                "name": "source-2",
                "srcset": [
                    ("1x", ["crop 100 100 200 200"]),
                    ("2x", ["crop 100 100 300 300"]),
                ],
            },
        ],
        "default": ("source-2", ["scale_in 800 600 True"]),
    },
}


@pytest.mark.parametrize(
    "orig_tag, new_tag, call_args",
    [
        (
            '<img class="image-process-thumb" src="/tmp/test.jpg" />',
            '<img class="image-process-thumb" src="/tmp/derivs/thumb/test.jpg"/>',
            [
                (
                    "tmp/test.jpg",
                    "tmp/derivs/thumb/test.jpg",
                    ["crop 0 0 50% 50%", "scale_out 150 150", "crop 0 0 150 150"],
                ),
            ],
        ),
        (
            '<img class="image-process-article-image" src="/tmp/test.jpg" />',
            '<img class="image-process-article-image" '
            'src="/tmp/derivs/article-image/test.jpg"/>',
            [
                (
                    "tmp/test.jpg",
                    "tmp/derivs/article-image/test.jpg",
                    ["scale_in 300 300"],
                ),
            ],
        ),
        (
            '<img class="image-process-crisp" src="/tmp/test.jpg" />',
            '<img class="image-process-crisp" src="/tmp/derivs/crisp/1x/test.jpg" '
            'srcset="/tmp/derivs/crisp/1x/test.jpg 1x, '
            "/tmp/derivs/crisp/2x/test.jpg 2x, "
            '/tmp/derivs/crisp/4x/test.jpg 4x"/>',
            [
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp/1x/test.jpg",
                    ["scale_in 800 600 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp/2x/test.jpg",
                    ["scale_in 1600 1200 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp/4x/test.jpg",
                    ["scale_in 3200 2400 True"],
                ),
            ],
        ),
        (
            '<img class="image-process-crisp2" src="/tmp/test.jpg" />',
            '<img class="image-process-crisp2" '
            'src="/tmp/derivs/crisp2/default/test.jpg" '
            'srcset="/tmp/derivs/crisp2/1x/test.jpg 1x, '
            "/tmp/derivs/crisp2/2x/test.jpg 2x, "
            '/tmp/derivs/crisp2/4x/test.jpg 4x"/>',
            # default must be first, because the process execute it first
            [
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp2/default/test.jpg",
                    ["scale_in 400 300 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp2/1x/test.jpg",
                    ["scale_in 800 600 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp2/2x/test.jpg",
                    ["scale_in 1600 1200 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/crisp2/4x/test.jpg",
                    ["scale_in 3200 2400 True"],
                ),
            ],
        ),
        (
            '<img class="image-process-large-photo" src="/tmp/test.jpg" />',
            '<img class="image-process-large-photo" '
            'sizes="(min-width: 1200px) 800px, (min-width: 992px) 650px, '
            '(min-width: 768px) 718px, 100vw" src="/tmp/derivs/large-photo/'
            '800w/test.jpg" '
            'srcset="/tmp/derivs/large-photo/600w/test.jpg 600w, '
            "/tmp/derivs/large-photo/800w/test.jpg 800w, "
            '/tmp/derivs/large-photo/1600w/test.jpg 1600w"/>',
            [
                (
                    "tmp/test.jpg",
                    "tmp/derivs/large-photo/600w/test.jpg",
                    ["scale_in 600 450 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/large-photo/800w/test.jpg",
                    ["scale_in 800 600 True"],
                ),
                (
                    "tmp/test.jpg",
                    "tmp/derivs/large-photo/1600w/test.jpg",
                    ["scale_in 1600 1200 True"],
                ),
            ],
        ),
        (
            '<picture><source class="source-1" '
            'src="/images/pelican-closeup.jpg"/><img '
            'class="image-process-pict" src="/images/pelican.jpg"/>'
            "</picture>",
            '<picture><source media="(min-width: 640px)" sizes="100vw" '
            'srcset="/images/derivs/pict/default/640w/pelican.jpg 640w, '
            "/images/derivs/pict/default/1024w/pelican.jpg 1024w, "
            '/images/derivs/pict/default/1600w/pelican.jpg 1600w"/>'
            '<source srcset="/images/derivs/pict/source-1/1x/'
            "pelican-closeup.jpg 1x, /images/derivs/pict/source-1/2x/"
            'pelican-closeup.jpg 2x"/><img '
            'class="image-process-pict" '
            'src="/images/derivs/pict/default/640w/pelican.jpg"/>'
            "</picture>",
            [
                (
                    "images/pelican.jpg",
                    "images/derivs/pict/default/640w/pelican.jpg",
                    ["scale_in 640 480 True"],
                ),
                (
                    "images/pelican.jpg",
                    "images/derivs/pict/default/1024w/pelican.jpg",
                    ["scale_in 1024 683 True"],
                ),
                (
                    "images/pelican.jpg",
                    "images/derivs/pict/default/1600w/pelican.jpg",
                    ["scale_in 1600 1200 True"],
                ),
                (
                    "images/pelican-closeup.jpg",
                    "images/derivs/pict/source-1/1x/pelican-closeup.jpg",
                    ["crop 100 100 200 200"],
                ),
                (
                    "images/pelican-closeup.jpg",
                    "images/derivs/pict/source-1/2x/pelican-closeup.jpg",
                    ["crop 100 100 300 300"],
                ),
            ],
        ),
        (
            '<div class="figure"><img alt="Pelican" class="image-process'
            '-pict2" src="/images/pelican.jpg" /> <p class="caption">'
            'A nice pelican</p> <div class="legend"> <img alt="Other view of '
            'pelican" class="image-process source-2" src="/images/'
            'pelican-closeup.jpg" /></div></div>',
            '<div class="figure"><picture><source media="(min-width: 640px)" '
            'sizes="100vw" srcset="/images/derivs/pict2/default/640w/pelican'
            ".jpg 640w, /images/derivs/pict2/default/1024w/pelican.jpg 1024w,"
            ' /images/derivs/pict2/default/1600w/pelican.jpg 1600w"/>'
            '<source srcset="/images/derivs/pict2/source-2/1x/pelican-closeup'
            ".jpg 1x, /images/derivs/pict2/source-2/2x/pelican-closeup.jpg "
            '2x"/><img alt="Pelican" class="image-process-pict2" '
            'src="/images/derivs/pict2/source-2/default/pelican-closeup.jpg"'
            '/></picture> <p class="caption">A nice pelican</p> <div '
            'class="legend"> </div></div>',
            [
                # Default calls first.
                (
                    "images/pelican-closeup.jpg",
                    "images/derivs/pict2/source-2/default/pelican-closeup.jpg",
                    ["scale_in 800 600 True"],
                ),
                (
                    "images/pelican.jpg",
                    "images/derivs/pict2/default/640w/pelican.jpg",
                    ["scale_in 640 480 True"],
                ),
                # Then images in processing order.
                (
                    "images/pelican.jpg",
                    "images/derivs/pict2/default/1024w/pelican.jpg",
                    ["scale_in 1024 683 True"],
                ),
                (
                    "images/pelican.jpg",
                    "images/derivs/pict2/default/1600w/pelican.jpg",
                    ["scale_in 1600 1200 True"],
                ),
                (
                    "images/pelican-closeup.jpg",
                    "images/derivs/pict2/source-2/1x/pelican-closeup.jpg",
                    ["crop 100 100 200 200"],
                ),
                (
                    "images/pelican-closeup.jpg",
                    "images/derivs/pict2/source-2/2x/pelican-closeup.jpg",
                    ["crop 100 100 300 300"],
                ),
            ],
        ),
    ],
)
def test_picture_generation(mocker, orig_tag, new_tag, call_args):
    # Allow non-existing images to be processed:
    mocker.patch(
        "pelican.plugins.image_process.image_process.is_img_identifiable",
        lambda img_filepath: True,
    )
    process = mocker.patch("pelican.plugins.image_process.image_process.process_image")

    settings = get_settings(
        IMAGE_PROCESS=COMPLEX_TRANSFORMS, IMAGE_PROCESS_DIR="derivs"
    )

    assert harvest_images_in_fragment(orig_tag, settings) == new_tag

    for i, (orig_img, new_img, transform_params) in enumerate(call_args):
        assert process.mock_calls[i] == mocker.call(
            (
                os.path.join(settings["PATH"], orig_img),
                os.path.join(settings["OUTPUT_PATH"], new_img),
                transform_params,
            ),
            settings,
        )


def process_image_mock_exif_tool_started(image, settings):
    assert ExifTool._instance is not None


def process_image_mock_exif_tool_not_started(image, settings):
    assert ExifTool._instance is None


@pytest.mark.parametrize("copy_tags", [True, False])
def test_exiftool_process_is_started_only_when_necessary(mocker, copy_tags):
    if shutil.which("exiftool") is None:
        warnings.warn(
            "EXIF tags copying will not be tested because the exiftool program could "
            "not be found. Please install exiftool and make sure it is in your path."
        )
        return

    if copy_tags:
        mocker.patch(
            "pelican.plugins.image_process.image_process.process_image",
            process_image_mock_exif_tool_started,
        )
    else:
        mocker.patch(
            "pelican.plugins.image_process.image_process.process_image",
            process_image_mock_exif_tool_not_started,
        )

    settings = get_settings(
        IMAGE_PROCESS_COPY_EXIF_TAGS=copy_tags,
        IMAGE_PROCESS=COMPLEX_TRANSFORMS,
        IMAGE_PROCESS_DIR="derivs",
    )

    harvest_images_in_fragment(
        '<img class="image-process-thumb" src="/tmp/test.jpg" />', settings
    )


@pytest.mark.parametrize("image_path", EXIF_TEST_IMAGES)
@pytest.mark.parametrize("copy_tags", [True, False])
def test_copy_exif_tags(tmp_path, image_path, copy_tags):
    if shutil.which("exiftool") is None:
        warnings.warn(
            "EXIF tags copying will not be tested because the exiftool program could "
            "not be found. Please install exiftool and make sure it is in your path."
        )
        return

    # A few EXIF tags to test for.
    exif_tags = [
        "Artist",
        "Creator",
        "Title",
        "Description",
        "Subject",
        "Rating",
        "ExifImageWidth",
    ]

    settings = get_settings(IMAGE_PROCESS_COPY_EXIF_TAGS=copy_tags)

    transform_id = "grayscale"
    transform_params = ["grayscale"]
    image_name = image_path.name
    destination_path = tmp_path.joinpath(transform_id, image_name)

    expected_results = subprocess.run(
        ["exiftool", "-json", image_path], stdout=subprocess.PIPE
    )
    expected_tags = json.loads(expected_results.stdout)[0]
    for tag in exif_tags:
        assert tag in expected_tags.keys()

    if copy_tags:
        ExifTool.start_exiftool()
    process_image((str(image_path), str(destination_path), transform_params), settings)
    if copy_tags:
        ExifTool.stop_exiftool()

    actual_results = subprocess.run(
        ["exiftool", "-json", destination_path], stdout=subprocess.PIPE
    )

    assert actual_results.returncode == 0

    actual_tags = json.loads(actual_results.stdout)[0]
    for tag in exif_tags:
        if copy_tags:
            assert tag in actual_tags.keys()
            assert expected_tags[tag] == actual_tags[tag]
        else:
            assert tag not in actual_tags.keys()


def test_is_img_identifiable():
    for test_image in TEST_IMAGES:
        assert is_img_identifiable(test_image)

    assert not is_img_identifiable("image/that/does/not/exist.png")

    assert not is_img_identifiable(TEST_DATA.joinpath("folded_puzzle.png"))
    assert not is_img_identifiable(TEST_DATA.joinpath("minimal.svg"))

    img = {"src": "https://upload.wikimedia.org/wikipedia/commons/3/34/Exemple.png"}
    settings = get_settings(IMAGE_PROCESS_DIR="derivatives")
    path = compute_paths(img, settings, derivative="thumb")
    assert not is_img_identifiable(path.source)


def generate_test_images():
    settings = get_settings()
    image_count = 0
    for image_path in TEST_IMAGES:
        for transform_id, transform_params in SINGLE_TRANSFORMS.items():
            process_image(
                (
                    str(image_path),
                    str(TRANSFORM_RESULTS.joinpath(transform_id, image_path.name)),
                    transform_params,
                ),
                settings,
            )
            image_count += 1

    print(f"{image_count} test images generated!")



"""
Image Process
=============

This plugin process images according to their class attribute.
"""
import codecs
import collections
import copy
import functools
import html
import logging
import os.path
import posixpath
import re
import shutil
import subprocess
import sys

from PIL import Image, ImageFilter, UnidentifiedImageError
from bs4 import BeautifulSoup
import six
from six.moves.urllib_parse import unquote, urlparse
from six.moves.urllib_request import pathname2url, url2pathname

from pelican import __version__ as pelican_version
from pelican import signals

log = logging.getLogger(__name__)

IMAGE_PROCESS_REGEX = re.compile("image-process-[-a-zA-Z0-9_]+")

Path = collections.namedtuple("Path", ["base_url", "source", "base_path", "filename"])


# A lot of inspiration from pyexiftool (https://github.com/smarnach/pyexiftool)
class ExifTool(object):
    errors = "strict"
    sentinel = b"{ready}"
    block_size = 4096

    _instance = None

    @staticmethod
    def start_exiftool():
        if shutil.which("exiftool") is None:
            log.warning(
                "[image_process] EXIF tags will not be copied because "
                "the exiftool program could not be found. "
                "Please install exiftool and make sure it is in your path."
            )
        else:
            ExifTool._instance = ExifTool()

    @staticmethod
    def copy_tags(src, dst):
        if ExifTool._instance is not None:
            ExifTool._instance._copy_tags(src, dst)

    @staticmethod
    def stop_exiftool():
        ExifTool._instance = None

    def __init__(self):
        self.encoding = sys.getfilesystemencoding()
        if self.encoding != "mbcs":
            try:
                codecs.lookup_error("surrogateescape")
            except LookupError:
                pass

        with open(os.devnull, "w") as devnull:
            self.process = subprocess.Popen(
                [
                    "exiftool",
                    "-stay_open",
                    "True",
                    "-@",
                    "-",
                    "-common_args",
                    "-G",
                    "-n",
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=devnull,
            )

    def __del__(self):
        if self.process is not None:
            self.process.terminate()

    def _copy_tags(self, src, dst):
        params = (
            b"-TagsFromFile",
            src.encode(self.encoding, ExifTool.errors),
            b'"-all:all>all:all"',
            dst.encode(self.encoding, ExifTool.errors),
        )
        self._send_command(params)
        params = (b"-delete_original!", dst.encode(self.encoding, ExifTool.errors))
        self._send_command(params)

    def _send_command(self, params):
        self.process.stdin.write(b"\n".join(params + (b"-j\n", b"-execute\n")))
        self.process.stdin.flush()
        output = b""
        fd = self.process.stdout.fileno()
        while not output.strip().endswith(ExifTool.sentinel):
            output += os.read(fd, ExifTool.block_size)
        exiftool_result = output.strip()[: -len(ExifTool.sentinel)]
        log.debug(
            "[image_process] exiftool result: {}".format(
                exiftool_result.decode("utf-8")
            )
        )


def convert_box(image, top, left, right, bottom):
    """Convert box coordinates strings to integer.

    t, l, r, b (top, left, right, bottom) must be strings specifying
    either a number or a percentage.
    """
    bbox = image.getbbox()
    img_width = bbox[2] - bbox[0]
    img_height = bbox[3] - bbox[1]

    if top[-1] == "%":
        top = img_height * float(top[:-1]) / 100.0
    else:
        top = float(top)
    if left[-1] == "%":
        left = img_width * float(left[:-1]) / 100.0
    else:
        left = float(left)
    if right[-1] == "%":
        right = img_width * float(right[:-1]) / 100.0
    else:
        right = float(right)
    if bottom[-1] == "%":
        bottom = img_height * float(bottom[:-1]) / 100.0
    else:
        bottom = float(bottom)

    return (top, left, right, bottom)


def crop(i, left, top, right, bottom):
    """Crop image i to the box (left, top)-(right, bottom).

    left, top, right, bottom must be strings specifying
    either a number or a percentage.
    """
    top, left, right, bottom = convert_box(i, top, left, right, bottom)
    return i.crop((int(left), int(top), int(right), int(bottom)))


def resize(i, w, h):
    """Resize the image to the dimension specified.

    w, h (width, height) must be strings specifying either a number
    or a percentage.
    """
    _, _, w, h = convert_box(i, "0", "0", w, h)

    if i.mode == "P":
        i = i.convert("RGBA")
    elif i.mode == "1":
        i = i.convert("L")

    return i.resize((int(w), int(h)), Image.Resampling.LANCZOS)


def scale(i, w, h, upscale, inside):
    """Resize the image to the dimension specified, keeping the aspect
    ratio.

    w, h (width, height) must be strings specifying either a number
    or a percentage, or "None" to ignore this constraint.

    If upscale is True, upscaling is allowed.

    If inside is True, the resulting image will not be larger than the
    dimensions specified, else it will not be smaller.
    """
    bbox = i.getbbox()
    iw = bbox[2] - bbox[0]
    ih = bbox[3] - bbox[1]

    if w == "None":
        w = 1.0
    elif w[-1] == "%":
        w = float(w[:-1]) / 100.0
    else:
        w = float(w) / iw

    if h == "None":
        h = 1.0
    elif h[-1] == "%":
        h = float(h[:-1]) / 100.0
    else:
        h = float(h) / ih

    if inside:
        scale = min(w, h)
    else:
        scale = max(w, h)

    if upscale in [0, "0", "False", False]:
        scale = min(scale, 1.0)

    if i.mode == "P":
        i = i.convert("RGBA")
    elif i.mode == "1":
        i = i.convert("L")

    return i.resize((int(scale * iw), int(scale * ih)), Image.Resampling.LANCZOS)


def rotate(i, degrees):
    if i.mode == "P":
        i = i.convert("RGBA")
    elif i.mode == "1":
        i = i.convert("L")

    # rotate does not support the LANCZOS filter (Pillow 2.7.0).
    return i.rotate(int(degrees), Image.Resampling.BICUBIC, True)


def apply_filter(i, f):
    if i.mode == "P":
        i = i.convert("RGBA")
    elif i.mode == "1":
        i = i.convert("L")

    return i.filter(f)


basic_ops = {
    "crop": crop,
    "flip_horizontal": lambda i: i.transpose(Image.Transpose.FLIP_LEFT_RIGHT),
    "flip_vertical": lambda i: i.transpose(Image.Transpose.FLIP_TOP_BOTTOM),
    "grayscale": lambda i: i.convert("L"),
    "resize": resize,
    "rotate": rotate,
    "scale_in": functools.partial(scale, inside=True),
    "scale_out": functools.partial(scale, inside=False),
    "blur": functools.partial(apply_filter, f=ImageFilter.BLUR),
    "contour": functools.partial(apply_filter, f=ImageFilter.CONTOUR),
    "detail": functools.partial(apply_filter, f=ImageFilter.DETAIL),
    "edge_enhance": functools.partial(apply_filter, f=ImageFilter.EDGE_ENHANCE),
    "edge_enhance_more": functools.partial(
        apply_filter, f=ImageFilter.EDGE_ENHANCE_MORE
    ),
    "emboss": functools.partial(apply_filter, f=ImageFilter.EMBOSS),
    "find_edges": functools.partial(apply_filter, f=ImageFilter.FIND_EDGES),
    "smooth": functools.partial(apply_filter, f=ImageFilter.SMOOTH),
    "smooth_more": functools.partial(apply_filter, f=ImageFilter.SMOOTH_MORE),
    "sharpen": functools.partial(apply_filter, f=ImageFilter.SHARPEN),
}


def harvest_images(path, context):
    log.debug("[image_process] harvesting %r", path)
    # Set default value for 'IMAGE_PROCESS_DIR'.
    if "IMAGE_PROCESS_DIR" not in context:
        context["IMAGE_PROCESS_DIR"] = "derivatives"

    # Set default value for 'IMAGE_PROCESS_ENCODING'
    if "IMAGE_PROCESS_ENCODING" not in context:
        context["IMAGE_PROCESS_ENCODING"] = "utf-8"

    # Set default value for 'IMAGE_PROCESS_COPY_EXIF_TAGS'
    if "IMAGE_PROCESS_COPY_EXIF_TAGS" not in context:
        context["IMAGE_PROCESS_COPY_EXIF_TAGS"] = False

    with open(path, "r+", encoding=context["IMAGE_PROCESS_ENCODING"]) as f:
        res = harvest_images_in_fragment(f, context)
        f.seek(0)
        f.truncate()
        f.write(res)


def harvest_feed_images(path, context, feed):
    # Set default value for 'IMAGE_PROCESS_DIR'.
    if "IMAGE_PROCESS_DIR" not in context:
        context["IMAGE_PROCESS_DIR"] = "derivatives"

    # Set default value for 'IMAGE_PROCESS_ENCODING'
    if "IMAGE_PROCESS_ENCODING" not in context:
        context["IMAGE_PROCESS_ENCODING"] = "utf-8"

    # Set default value for 'IMAGE_PROCESS_COPY_EXIF_TAGS'
    if "IMAGE_PROCESS_COPY_EXIF_TAGS" not in context:
        context["IMAGE_PROCESS_COPY_EXIF_TAGS"] = False

    with open(path, "r+", encoding=context["IMAGE_PROCESS_ENCODING"]) as f:
        soup = BeautifulSoup(f, "xml")

        for content in soup.find_all("content"):
            if content["type"] != "html":
                continue

            doc = html.unescape(content.string)
            res = harvest_images_in_fragment(doc, context)
            content.string = res
        f.seek(0)
        f.truncate()
        f.write(str(soup))


def harvest_images_in_fragment(fragment, settings):
    parser = settings.get("IMAGE_PROCESS_PARSER", "html.parser")
    soup = BeautifulSoup(fragment, parser)

    copy_exif_tags = settings.get("IMAGE_PROCESS_COPY_EXIF_TAGS", False)
    if copy_exif_tags:
        ExifTool.start_exiftool()

    for img in soup.find_all("img", class_=IMAGE_PROCESS_REGEX):
        for c in img["class"]:
            if c.startswith("image-process-"):
                derivative = c[14:]
                break
        else:
            continue

        try:
            d = settings["IMAGE_PROCESS"][derivative]
        except KeyError:
            raise RuntimeError("Derivative %s undefined." % derivative)

        if isinstance(d, list):
            # Single source image specification.
            process_img_tag(img, settings, derivative)

        elif not isinstance(d, dict):
            raise RuntimeError(
                "Derivative %s definition not handled"
                "(must be list or dict)" % (derivative)
            )

        elif "type" not in d:
            raise RuntimeError('"type" is mandatory for %s.' % derivative)

        elif d["type"] == "image":
            # Single source image specification.
            process_img_tag(img, settings, derivative)

        elif d["type"] == "responsive-image" and "srcset" not in img.attrs:
            # srcset image specification.
            build_srcset(img, settings, derivative)

        elif d["type"] == "picture":
            # Multiple source (picture) specification.
            group = img.find_parent()
            if group.name == "div":
                convert_div_to_picture_tag(soup, img, group, settings, derivative)
            elif group.name == "picture":
                process_picture(soup, img, group, settings, derivative)

    ExifTool.stop_exiftool()
    return str(soup)


def compute_paths(img, settings, derivative):
    process_dir = settings["IMAGE_PROCESS_DIR"]
    img_src = urlparse(img["src"])
    img_src_path = url2pathname(img_src.path.lstrip("/"))
    img_src_dirname, filename = os.path.split(img_src_path)
    derivative_path = os.path.join(process_dir, derivative)
    # urljoin truncates leading ../ elements
    base_url = posixpath.join(
        posixpath.dirname(img["src"]), pathname2url(derivative_path)
    )

    if pelican_version != "unknown" and int(pelican_version.split(".")[0]) < 4:
        file_paths = settings["filenames"]
    else:
        file_paths = settings["static_content"]

    for f, contobj in file_paths.items():
        save_as = contobj.get_url_setting("save_as")
        # save_as can be set to empty string, which would match everything
        if save_as and img_src_path.endswith(save_as):
            source = contobj.source_path
            base_path = os.path.join(
                contobj.settings["OUTPUT_PATH"],
                os.path.dirname(contobj.get_url_setting("save_as")),
                process_dir,
                derivative,
            )
            break
    else:
        if "SITEURL" in settings:
            site_url = urlparse(settings["SITEURL"])
            site_url_path = url2pathname(site_url.path[1:])
        else:
            # if SITEURL is undefined, don't break!
            site_url_path = None

        if site_url_path:
            src_path = img_src_path.partition(site_url_path)[2].lstrip("/")
        else:
            src_path = img_src_path.lstrip("/")
        source = os.path.join(settings["PATH"], src_path)
        base_path = os.path.join(
            settings["OUTPUT_PATH"], os.path.dirname(src_path), derivative_path
        )

    return Path(base_url, source, base_path, filename)


def process_img_tag(img, settings, derivative):
    path = compute_paths(img, settings, derivative)
    if not is_img_identifiable(path.source):
        log.warn(
            "[image_process] Skipping image %s that could not be identified by Pillow",
            path.source,
        )
        return
    process = settings["IMAGE_PROCESS"][derivative]

    img["src"] = posixpath.join(path.base_url, path.filename)
    destination = os.path.join(path.base_path, path.filename)

    if not isinstance(process, list):
        process = process["ops"]

    process_image((path.source, destination, process), settings)


def is_img_identifiable(img_filepath):
    try:
        Image.open(img_filepath)
        return True
    except (FileNotFoundError, UnidentifiedImageError):
        return False


def build_srcset(img, settings, derivative):
    path = compute_paths(img, settings, derivative)
    if not is_img_identifiable(path.source):
        log.warn(
            "[image_process] Skipping image %s that could not be identified by Pillow",
            path.source,
        )
        return
    process = settings["IMAGE_PROCESS"][derivative]

    default = process["default"]
    if isinstance(default, six.string_types):
        breakpoints = {i for i, _ in process["srcset"]}
        if default not in breakpoints:
            log.error(
                '[image_process] srcset "%s" does not define default "%s"',
                derivative,
                default,
            )
        default_name = default
    elif isinstance(default, list):
        default_name = "default"
        destination = os.path.join(path.base_path, default_name, path.filename)
        process_image((path.source, destination, default), settings)

    img["src"] = posixpath.join(path.base_url, default_name, path.filename)

    if "sizes" in process:
        img["sizes"] = process["sizes"]

    srcset = []
    for src in process["srcset"]:
        file_path = posixpath.join(path.base_url, src[0], path.filename)
        srcset.append("%s %s" % (file_path, src[0]))
        destination = os.path.join(path.base_path, src[0], path.filename)
        process_image((path.source, destination, src[1]), settings)

    if len(srcset) > 0:
        img["srcset"] = ", ".join(srcset)


def convert_div_to_picture_tag(soup, img, group, settings, derivative):
    """
    Convert a div containing multiple images to a picture.
    """
    process_dir = settings["IMAGE_PROCESS_DIR"]
    # Compile sources URL. Special source "default" uses the main
    # image URL. Other sources use the img with classes
    # [source['name'], 'image-process'].  We also remove the img from
    # the DOM.
    sources = copy.deepcopy(settings["IMAGE_PROCESS"][derivative]["sources"])
    for s in sources:
        if s["name"] == "default":
            s["url"] = img["src"]
        else:
            candidates = group.find_all("img", class_=s["name"])
            for candidate in candidates:
                if "image-process" in candidate["class"]:
                    s["url"] = candidate["src"]
                    candidate.decompose()
                    break

        url_path, s["filename"] = os.path.split(s["url"])
        s["base_url"] = os.path.join(url_path, process_dir, derivative)
        s["base_path"] = os.path.join(settings["OUTPUT_PATH"], s["base_url"][1:])

    # If default is not None, change default img source to the image
    # derivative referenced.
    default = settings["IMAGE_PROCESS"][derivative]["default"]
    if default is not None:
        default_source_name = default[0]

        default_source = None
        for s in sources:
            if s["name"] == default_source_name:
                default_source = s
                break

        if default_source is None:
            raise RuntimeError(
                'No source matching "%s", referenced in default setting.',
                (default_source_name,),
            )

        if isinstance(default[1], six.string_types):
            default_item_name = default[1]

        elif isinstance(default[1], list):
            default_item_name = "default"

            source = os.path.join(settings["PATH"], default_source["url"][1:])
            destination = os.path.join(
                s["base_path"],
                default_source_name,
                default_item_name,
                default_source["filename"],
            )
            process_image((source, destination, default[1]), settings)

        # Change img src to url of default processed image.
        img["src"] = os.path.join(
            s["base_url"],
            default_source_name,
            default_item_name,
            default_source["filename"],
        )

    # Create picture tag.
    picture_tag = soup.new_tag("picture")
    for s in sources:
        # Create new <source>
        source_attrs = {k: s[k] for k in s if k in ["media", "sizes"]}
        source_tag = soup.new_tag("source", **source_attrs)

        srcset = []
        for src in s["srcset"]:
            srcset.append(
                "%s %s"
                % (
                    os.path.join(s["base_url"], s["name"], src[0], s["filename"]),
                    src[0],
                )
            )

            source = os.path.join(settings["PATH"], s["url"][1:])
            destination = os.path.join(s["base_path"], s["name"], src[0], s["filename"])
            process_image((source, destination, src[1]), settings)

        if len(srcset) > 0:
            source_tag["srcset"] = ", ".join(srcset)

        picture_tag.append(source_tag)

    # Wrap img with <picture>
    img.wrap(picture_tag)


def process_picture(soup, img, group, settings, derivative):
    """
    Convert a simplified picture to a full HTML picture:

    <picture>
        <source class="source-1" src="image1.jpg"></source>
        <source class="source-2" src="image2.jpg"></source>
        <img class="image-process-picture" src="image3.jpg"></img>
    </picture>

    to

    <picture>
        <source srcset="...image1.jpg..." media="..." sizes="..."></source>
        <source srcset="...image2.jpg..."></source>
        <source srcset="...image3.jpg..." media="..." sizes="..."></source>
        <img src=".../image3.jpg"></img>
    </picture>

    """

    process_dir = settings["IMAGE_PROCESS_DIR"]
    process = settings["IMAGE_PROCESS"][derivative]
    # Compile sources URL. Special source "default" uses the main
    # image URL. Other sources use the <source> with classes
    # source['name'].  We also remove the <source>s from the DOM.
    sources = copy.deepcopy(process["sources"])
    for s in sources:
        if s["name"] == "default":
            s["url"] = img["src"]
            source_attrs = {k: s[k] for k in s if k in ["media", "sizes"]}
            s["element"] = soup.new_tag("source", **source_attrs)
        else:
            s["element"] = group.find("source", class_=s["name"]).extract()
            s["url"] = s["element"]["src"]
            del s["element"]["src"]
            del s["element"]["class"]

        url_path, s["filename"] = os.path.split(s["url"])
        s["base_url"] = posixpath.join(url_path, process_dir, derivative)
        s["base_path"] = os.path.join(settings["OUTPUT_PATH"], s["base_url"][1:])

    # If default is not None, change default img source to the image
    # derivative referenced.
    default = process["default"]
    if default is not None:
        default_source_name = default[0]

        default_source = None
        for s in sources:
            if s["name"] == default_source_name:
                default_source = s
                break

        if default_source is None:
            raise RuntimeError(
                'No source matching "%s", referenced in default setting.',
                (default_source_name,),
            )

        if isinstance(default[1], six.string_types):
            default_item_name = default[1]

        elif isinstance(default[1], list):
            default_item_name = "default"
            source = os.path.join(settings["PATH"], default_source["url"][1:])
            destination = os.path.join(
                s["base_path"],
                default_source_name,
                default_item_name,
                default_source["filename"],
            )

            process_image((source, destination, default[1]), settings)

        # Change img src to url of default processed image.
        img["src"] = posixpath.join(
            s["base_url"],
            default_source_name,
            default_item_name,
            default_source["filename"],
        )

    # Generate srcsets and put back <source>s in <picture>.
    for s in sources:
        srcset = []
        for src in s["srcset"]:
            srcset.append(
                "%s %s"
                % (
                    posixpath.join(s["base_url"], s["name"], src[0], s["filename"]),
                    src[0],
                )
            )

            source = os.path.join(settings["PATH"], s["url"][1:])
            destination = os.path.join(s["base_path"], s["name"], src[0], s["filename"])
            process_image((source, destination, src[1]), settings)

        if len(srcset) > 0:
            # Append source elements to the picture in the same order
            # as they are found in
            # settings['IMAGE_PROCESS'][derivative]['sources'].
            s["element"]["srcset"] = ", ".join(srcset)
            img.insert_before(s["element"])


def process_image(image, settings):
    # Set default value for 'IMAGE_PROCESS_FORCE'.
    if "IMAGE_PROCESS_FORCE" not in settings:
        settings["IMAGE_PROCESS_FORCE"] = False

    # remove URL encoding to get to physical filenames
    image = list(image)
    image[0] = unquote(image[0])
    image[1] = unquote(image[1])
    # image[2] is the transformation

    log.debug("[image_process] {} -> {}".format(image[0], image[1]))

    os.makedirs(os.path.dirname(image[1]), exist_ok=True)

    # If original image is older than existing derivative, skip
    # processing to save time, unless user explicitly forced
    # image generation.
    if (
        settings["IMAGE_PROCESS_FORCE"]
        or not os.path.exists(image[1])
        or os.path.getmtime(image[0]) > os.path.getmtime(image[1])
    ):

        i = Image.open(image[0])

        for step in image[2]:
            if hasattr(step, "__call__"):
                i = step(i)
            else:
                elems = step.split(" ")
                i = basic_ops[elems[0]](i, *(elems[1:]))

        # `save_all=True`  will allow saving multi-page (aka animated) GIF's
        # however, turning it on seems to break PNG support, and doesn't seem
        # to work on GIF's either...
        i.save(image[1], progressive=True)

        ExifTool.copy_tags(image[0], image[1])


def register():
    signals.content_written.connect(harvest_images)
    signals.feed_written.connect(harvest_feed_images)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pelican', 'pelican.plugins.image_process']

package_data = \
{'': ['*'],
 'pelican.plugins.image_process': ['test_data/*',
                                   'test_data/exif/*',
                                   'test_data/results/blur/*',
                                   'test_data/results/contour/*',
                                   'test_data/results/crop/*',
                                   'test_data/results/detail/*',
                                   'test_data/results/edge_enhance/*',
                                   'test_data/results/edge_enhance_more/*',
                                   'test_data/results/emboss/*',
                                   'test_data/results/find_edges/*',
                                   'test_data/results/flip_horizontal/*',
                                   'test_data/results/flip_vertical/*',
                                   'test_data/results/grayscale/*',
                                   'test_data/results/resize/*',
                                   'test_data/results/rotate/*',
                                   'test_data/results/scale_in/*',
                                   'test_data/results/scale_out/*',
                                   'test_data/results/sharpen/*',
                                   'test_data/results/smooth/*',
                                   'test_data/results/smooth_more/*']}

install_requires = \
['Pillow>=9.0,<10.0',
 'beautifulsoup4>=4.9,<5.0',
 'lxml>=4.6,<5.0',
 'pelican>=3,<5',
 'six>=1.15.0,<2.0.0']

extras_require = \
{'markdown': ['markdown>=3.2,<4.0']}

setup_kwargs = {
    'name': 'pelican-image-process',
    'version': '3.0.1',
    'description': 'Pelican plugin that automates image processing.',
    'long_description': '# Image Process: A Plugin for Pelican\n\n[![Build Status](https://img.shields.io/github/workflow/status/pelican-plugins/image-process/build)](https://github.com/pelican-plugins/image-process/actions)\n[![PyPI Version](https://img.shields.io/pypi/v/pelican-image-process)](https://pypi.org/project/pelican-image-process/)\n![License](https://img.shields.io/badge/license-AGPL--3.0-blue)\n\n*Image Process* is a plugin for [Pelican](https://getpelican.com),\na static site generator written in Python.\n\n*Image Process* let you automate the processing of images based on their\nclass attribute. Use this plugin to minimize the overall page weight\nand to save you a trip to Gimp or Photoshop each time you include an\nimage in your post.\n\n*Image Process* also makes it easy to create responsive images using\nthe HTML5 `srcset` attribute and `<picture>` tag. It does this\nby generating multiple derivative images from one or more sources.\n\n*Image Process* will not overwrite your original images.\n\n## Installation\n\nThe easiest way to install *Image Process* is via Pip. This\nwill also install the required dependencies automatically.\n\n```sh\npython -m pip install pelican-image-process\n```\n\nYou will then need to configure your desired transformations (see *Usage*\nbelow) and add the appropriate class to images you want processed.\n\n## Usage\n\n*Image Process* scans your content for `<img>` tags with special\nclasses. It then maps the classes to a set of image processing\ninstructions, computes new images, and modifies HTML code according to\nthe instructions.\n\n### Define Transformations\n\nThe first step in using this module is to define some image\ntransformations in your Pelican configuration file. Transformations\nare defined in the `IMAGE_PROCESS` dictionary, mapping a\ntransformation name to a list of operations. There are three kinds of\ntransformations: image replacement, responsive image, and picture set.\n\n#### Image Replacement\n\nThe simplest image processing will replace the original image by a\nnew, transformed image computed from the original. You may use this,\nfor example, to ensure that all images are of the same size, or to\ncompute a thumbnail from a larger image:\n\n```python\nIMAGE_PROCESS = {\n    "article-image": ["scale_in 300 300 True"],\n    "thumb": ["crop 0 0 50% 50%", "scale_out 150 150 True", "crop 0 0 150 150"],\n}\n```\n\nThese transformations tell *Image Process* to transform the image\nreferred to by the `src` attribute of an `<img>` according to the\nlist of operations specified, and replace the `src` attribute with the\nURL of the transformed image.\n\nFor consistency with other types of transformations described\nbelow, there is an alternative syntax for the processing instructions:\n\n```python\nIMAGE_PROCESS = {\n    "thumb": {\n        "type": "image",\n        "ops": ["crop 0 0 50% 50%", "scale_out 150 150 True", "crop 0 0 150 150"],\n    },\n    "article-image": {\n        "type": "image",\n        "ops": ["scale_in 300 300 True"],\n    },\n}\n```\n\nTo apply image replacement to the images in your articles, you must add to them\nthe special class `image-process-<transform>`, in which `<transform>` is the ID\nof the transformation you wish to apply.\n\nLet\'s say you have defined the transformation described above. To get your\nimage processed, it needs to have the right CSS class:\n\n```html\n<img class="image-process-article-image" src="/images/pelican.jpg"/>\n```\n\nThis can be produced in Markdown with:\n\n```markdown\n![](/images/pelican.png){: .image-process-article-image}\n```\n\nIn reStructuredText, use the `:class:` attribute of the `image` or\nthe `figure` directive:\n\n```rst\n.. image:: /images/pelican.png\n   :class: image-process-article-image\n```\n\n```rst\n.. figure:: /images/pelican.png\n    :class: image-process-article-image\n```\n\n⚠️ **Warning:**\n\n> The reStructuredText reader will convert underscores (`_`) to\n> dashes (`-`) in class names. To make sure everything runs\n> smoothly, do not use underscores in your transformation names.\n\n\n#### Responsive Images\n\nYou can use *Image Process* to automatically generate a set of\nimages that will be selected for display by browsers according to the\nviewport width or according to the device resolution. To accomplish\nthis, *Image Process* will add a [`srcset` attribute](https://caniuse.com/srcset)\n(and maybe a `media` and a `sizes` attribute) to the `<img>` tag.\n\nHTML5 supports two types of responsive image sets. The first one is\ndevice-pixel-ratio-based, selecting higher resolution images for higher\nresolution devices; the second one is viewport-based, selecting\nimages according to the viewport size. You can read more about\n[HTML5 responsive images][] for a gentle introduction to the `srcset`\nand `<picture>` syntaxes.\n\nTo tell *Image Process* to generate a responsive image, add a\n`responsive-image` transformation to your your `IMAGE_PROCESS`\ndictionary, with the following syntax:\n\n```python\nIMAGE_PROCESS = {\n    "crisp": {\n        "type": "responsive-image",\n        "srcset": [\n            ("1x", ["scale_in 800 600 True"]),\n            ("2x", ["scale_in 1600 1200 True"]),\n            ("4x", ["scale_in 3200 2400 True"]),\n        ],\n        "default": "1x",\n    },\n    "large-photo": {\n        "type": "responsive-image",\n        "sizes": (\n            "(min-width: 1200px) 800px, "\n            "(min-width: 992px) 650px, "\n            "(min-width: 768px) 718px, "\n            "100vw"\n        ),\n        "srcset": [\n            ("600w", ["scale_in 600 450 True"]),\n            ("800w", ["scale_in 800 600 True"]),\n            ("1600w", ["scale_in 1600 1200 True"]),\n        ],\n        "default": "800w",\n    },\n}\n```\n\nThe `crisp` transformation is an example of a transformation\nenabling device-pixel-ratio-based selection. The `srcset` is a list\nof tuples, each tuple containing the image description (`"1x"`,\n`"2x"`, etc.) and the list of operations to generate the derivative\nimage from the original image (the original image is the value of the\n`src` attribute of the `<img>`). Image descriptions are hints\nabout the resolution of the associated image and must have the suffix\n`x`. The `default` setting specifies the image to use to replace the `src`\nattribute of the image.  This is the image that will be displayed by\nbrowsers that do not support the `srcset` syntax.\n\nThe `large-photo` transformation is an example of a transformation\nenabling viewport-based selection. The `sizes` contains a rule to\ncompute the width of the displayed image from the width of the\nviewport. Once the browser knows the image width, it will select an\nimage source from the `srcset`. The `srcset` is a list of tuple,\neach tuple containing the image description (`"600w"`, `"800w"`,\netc.) and the list of operations to generate the derivative image from\nthe original image (the original image is the value of the `src`\nattribute of the `<img>`). Image descriptions are hints about the\nwidth in pixels of the associated image and must have the suffix\n`w`. The `default` setting specifies the image to use to replace the `src`\nattribute of the image.  This is the image that will be displayed by\nbrowsers that do not support the `srcset` syntax.\n\nIn the two examples above, the `default` setting is a string referring to\none of the images in the `srcset`. However, the `default` value\ncould also be a list of operations to generate a different derivative\nimage.\n\nTo make the images in your article responsive, you must add to them the\nspecial class `image-process-<transform>`, in which `<transform>` is the ID of the\ntransformation you wish to apply, exactly like you would do for the\nimage replacement case, described above.\n\nSo, in HTML it should look like this:\n\n```html\n<img class="image-process-large-photo" src="/images/pelican.jpg"/>\n```\n\nWhich can be produced in Markdown with:\n\n```markdown\n![](/images/pelican.jpg){: .image-process-large-photo}\n```\n\nIn reStructuredText, use the `:class:` attribute of the `image` or\nthe `figure` directive:\n\n```rst\n.. image:: /images/pelican.jpg\n   :class: image-process-large-photo\n```\n\n```rst\n.. figure:: /images/pelican.jpg\n    :class: image-process-large-photo\n```\n\n#### Picture Set\n\n*Image Process* can be used to generate the images used by a\n`<picture>` tag. The `<picture>` syntax allows for more\nflexibility in providing a choice of image to the browser.\nAgain, you can read more about [HTML5 responsive images][] for a\ngentle introduction to the `srcset` and `<picture>` syntaxes.\n\nTo tell *Image Process* to generate the images for a `<picture>`,\nadd a `picture` entry to your `IMAGE_PROCESS` dictionary with the\nfollowing syntax:\n\n```python\nIMAGE_PROCESS = {\n    "example-pict": {\n        "type": "picture",\n        "sources": [\n            {\n                "name": "default",\n                "media": "(min-width: 640px)",\n                "srcset": [\n                    ("640w", ["scale_in 640 480 True"]),\n                    ("1024w", ["scale_in 1024 683 True"]),\n                    ("1600w", ["scale_in 1600 1200 True"]),\n                ],\n                "sizes": "100vw",\n            },\n            {\n                "name": "source-1",\n                "srcset": [\n                    ("1x", ["crop 100 100 200 200"]),\n                    ("2x", ["crop 100 100 300 300"]),\n                ]\n            },\n        ],\n        "default": ("default", "640w"),\n    },\n}\n```\n\nEach of the `sources` entries is very similar to the `responsive\nimage` describe above. Here, each source must have a `name`, which\nwill be used to find the URL of the original image for this source in\nyour article. The source may also have a `media`, which contains a\nrule used by the browser to select the active source. The `default`\nsetting specifies the image to use to replace the `src` attribute of\nthe `<img>` inside the `<picture>`.  This is the image that will be\ndisplayed by browsers that do not support the `<picture>` syntax. In\nthis example, it will use the image `640w` from the source `default`.\nA list of operations could have been specified instead of `640w`.\n\nTo generate a responsive `<picture>` for the images in your\narticles, you must add to your article a pseudo `<picture>` tag that\nlooks like this:\n\n```html\n<picture>\n    <source class="source-1" src="/images/pelican-closeup.jpg"></source>\n    <img class="image-process-example-pict" src="/images/pelican.jpg"/>\n</picture>\n```\n\nEach `<source>` tag maps a source name (the `class` attribute) to\na file (the `src` attribute). The `<img>` must have the special\nclass `image-process-` followed by the name of the transformation\nyou wish to apply. The file referenced by the `src` attribute of the\n`<img>` will be used as the special `default` source in your\ntransformation definition.\n\nYou can\'t produce this with pure Markdown and must instead resort to raw HTML.\n\nIn reStructuredText, however, you can also use the `figure` directive\nto generate a `<picture>`. The figure image file will be used as the\nspecial `default` source; other sources must be added in the legend\nsection of the `figure` as `image` directives. The figure class must\nbe `image-process-` followed by the name of the transformation you\nwish to apply, while the other images must have two classes:\n`image-process` and the name of the source they provide an image for:\n\n```rst\n.. figure:: /images/pelican.jpg\n   :class: image-process-example-pict\n\n    Test picture\n\n    .. image:: /images/pelican-closeup.jpg\n       :class: image-process source-1\n```\n\nThe images in the legend section that are used as source for the\n`<picture>` will be removed from the image legend, so that they do\nnot appear in your final article.\n\n### Transformations\n\nAvailable operations for transformations are:\n\n* `crop <top> <left> <right> <bottom>`:\n\n    Crop the image to the box (`<left>`, `<top>`)-(`<right>`, `<bottom>`). Values\n    can be absolute (a number) or relative to the size of the image (a\n    number followed by a percent sign `%`).\n\n* `flip_horizontal`:\n\n    Flip the image horizontally.\n\n* `flip_vertical`:\n\n    Flip the image vertically.\n\n* `grayscale`:\n\n    Convert the image to grayscale.\n\n* `resize <width> <height>`:\n\n    Resize the image. This operation does *not* preserve the image aspect\n    ratio. Values can be absolute (a number) or relative to the\n    size of the image (a number followed by a percent sign `%`).\n\n* `rotate <degrees>`:\n\n    Rotate the image.\n\n* `scale_in <width> <height> <upscale>`:\n\n    Resize the image. This operation preserves the image aspect ratio\n    and the resulting image will be no larger than `<width>` x\n    `<height>`. Values can be absolute (a number) or relative to the\n    size of the image (a number followed by a percent sign `%`).\n    If `<upscale>` is `False`, smaller images will not be enlarged.\n\n* `scale_out <width> <height> <upscale>`:\n\n    Resize the image. This operation preserves the image aspect ratio\n    and the resulting image will be no smaller than `<width>` x\n    `<height>`. Values can be absolute (a number) or relative to the\n    size of the image (a number followed by a percent sign `%`).\n    If `<upscale>` is `False`, smaller images will not be enlarged.\n\n* `blur`:\n\n    Apply the `pillow.ImageFilter.BLUR` filter to the image.\n\n* `contour`:\n\n    Apply the `pillow.ImageFilter.CONTOUR` filter to the image.\n\n* `detail`:\n\n    Apply the `pillow.ImageFilter.DETAIL` filter to the image.\n\n* `edge_enhance`:\n\n    Apply the `pillow.ImageFilter.EDGE_ENHANCE` filter to the image.\n\n* `edge_enhance_more`:\n\n    Apply the `pillow.ImageFilter.EDGE_ENHANCE_MORE` filter to the image.\n\n* `emboss`:\n\n    Apply the `pillow.ImageFilter.EMBOSS` filter to the image.\n\n* `find_edges`:\n\n    Apply the `pillow.ImageFilter.FIND_EDGES` filter to the image.\n\n* `smooth`:\n\n    Apply the `pillow.ImageFilter.SMOOTH filter` to the image.\n\n* `smooth_more`:\n\n    Apply the `pillow.ImageFilter.SMOOTH_MORE` filter to the image.\n\n* `sharpen`:\n\n    Apply the `pillow.ImageFilter.SHARPEN` filter to the image.\n\nYou can also define your own operations; the only requirement is that\nyour operation should be a callable object expecting a `pillow.Image` as\nits first parameter and it should return the transformed image:\n\n```python\ndef crop_face(image):\n    """Detect face in image and crop around it."""\n    # Fancy algorithm.\n    return image\n\nIMAGE_PROCESS = {\n    "face-thumbnail": [crop_face, "scale_out 150 150 True"]\n}\n```\n\n### Additional Settings\n\n#### Destination Directory\n\nBy default, the new images will be stored in a directory named\n`derivative/<TRANSFORMATION_NAME>` in the output folder at\nthe same location as the original image.\nFor example, if the original image is located in\nthe `content/images` folder, the computed images will be stored\nin `output/images/derivative/<TRANSFORMATION_NAME>`.\nAll the transformations are done in the output directory in order\nto avoid confusion with the source files or if we test multiple\ntransformations. You can replace `derivative` by something else using\nthe `IMAGE_PROCESS_DIR` setting in your Pelican settings file:\n\n```python\nIMAGE_PROCESS_DIR = "derivees"\n```\n\n#### Force Image Processing\n\nIf the transformed image already exists and is newer than the original\nimage, the plugin assumes that it should not re-compute it again. You\ncan force the plugin to re-compute all images by setting\n`IMAGE_PROCESS_FORCE` to `True` in your Pelican configuration file.\n\n```python\nIMAGE_PROCESS_FORCE = True\n```\n\n#### Selecting a HTML Parser\n\nYou may select the HTML parser which is used. The default is the built-in\n`html.parser` but you may also select `html5lib` or `lxml` by setting\n`IMAGE_PROCESS_PARSER` in your Pelican settings file. For example:\n\n```python\nIMAGE_PROCESS_PARSER = "html5lib"\n```\n\nFor details, refer to the [BeautifulSoup documentation on parsers][].\n\n#### File Encoding\n\nYou may select a different file encoding to be used by BeautifulSoup as it\nopens your files. The default is `utf-8`.\n\n```python\nIMAGE_PROCESS_ENCODING = "utf-8"\n```\n\n#### Copying EXIF Tags\n\nYou may ask _Image Process_ to copy the EXIF tags from your original image to\nthe transformed images. You must have [exiftool](https://exiftool.org/) installed.\n\n```python\nIMAGE_PROCESS_COPY_EXIF_TAGS = True\n```\n\n## Known Issues\n\n* Pillow, when resizing animated GIF files, [does not return an animated file](https://github.com/pelican-plugins/image-process/issues/11).\n\n## Contributing\n\nContributions are welcome and much appreciated. Every little bit helps. You can contribute by improving the documentation, adding missing features, and fixing bugs. You can also help out by reviewing and commenting on [existing issues][].\n\nTo start contributing to this plugin, review the [Contributing to Pelican][] documentation, beginning with the **Contributing Code** section.\n\n[existing issues]: https://github.com/pelican-plugins/image-process/issues\n[Contributing to Pelican]: https://docs.getpelican.com/en/latest/contribute.html\n\n### Regenerating Test Images\n\nIf you need to regenerate the transformed images used by the test suite, there\nis a helper function to do this for you. From the Python REPL:\n\n```pycon\n>>> from pelican.plugins.image_process.test_image_process import generate_test_images\n>>> generate_test_images()\n36 test images generated!\n```\n\n## License\n\nThis project is licensed under the [AGPL-3.0 license](http://www.gnu.org/licenses/agpl-3.0.html).\n\n[Pelican image](https://web.archive.org/web/20090505115626/http://www.pdphoto.org/PictureDetail.php?mat=&pg=5726) in test data by Jon Sullivan. Published under a [Creative Commons Public Domain license](https://creativecommons.org/licenses/publicdomain/).\n\n\n[HTML5 responsive images]: https://www.smashingmagazine.com/2014/05/14/responsive-images-done-right-guide-picture-srcset/\n[BeautifulSoup documentation on parsers]: https://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser\n',
    'author': 'Pelican Dev Team',
    'author_email': 'authors@getpelican.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pelican-plugins/image-process',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
